// The DynoIndexer reflection sample.
//
// The DynoIndexer class provided below accepts any object
// reference as a constructor argument and the proceeds
// to provide indexer-based access to all of the fields
// and properties of the target object.
//
// The getter can be used to fetch the value of the same-named
// field or property.  The setter can be used to modify any
// matching field, or invoke any matching setter on a property.
//
// Although of questionable value (but interesting nonetheless :-),
// the indexer also provides access methods via the indexer
// in the following manner:
//
// When the getter is invoked, the indexer will invoke
// a method matching the specified member name as long as
// the method takes no arguments and has a non-void return value.
//
// When the setter is invoked, the indexer will invoke
// a method matching the specified member name as long as
// the method takes only one argument, and that arguments type
// exactly matches the type of the value being passed in
// (no type coercion is performed - that's left as an exercise
// for the reader).
//
// Also left as an exercise for the reader is modifying the
// indexer so that type information could be passed that would
// allow the DynoIndexer to disambiguate between overloaded
// method names.
//
// Mike Woodring
// http://www.bearcanyon.com
// http://staff.develop.com/woodring
//
using System;
using System.Reflection;

class Point
{
    public Point( int x, int y )
    {
        this.x = x;
        this.y = y;
    }

    private int x;
    private int y;
    
    public int xPos
    {
        get { return(x); }
        set { x = value; }
    }

    public int yPos
    {
        get { return(y); }
        set { y = value; }
    }

    public int Sum()
    {
        return(x + y);
    }

    public void Scale( int factor )
    {
        x *= factor;
        y *= factor;
    }
}

class Test
{
    static void Main()
    {
        Point p = new Point(5, 10);
        DynoIndexer di = new DynoIndexer(p, true);

        // Test the setter.
        //
        di["x"] = 10;
        di["y"] = 20;
        di["xPos"] = 30;
        di["yPos"] = 40;
        di["Scale"] = -2;

        // Test the getter.
        //
        Console.WriteLine("x    = {0}", di["x"]);
        Console.WriteLine("y    = {0}", di["y"]);
        Console.WriteLine("xPos = {0}", di["xPos"]);
        Console.WriteLine("yPos = {0}", di["yPos"]);
        Console.WriteLine("Sum  = {0}", di["Sum"]);
    }
}

class DynoIndexer
{
    private object          target;
    private BindingFlags    bindingFlags;
    private Type            targetType;

    // This constructor only attempts to operate on the
    // public members of the underlying target object.
    //
    public DynoIndexer( object target )
        : this(target, false)
    {
    }

    // This constructor allows the caller to attempt access
    // to public as well as non-public members of the underlying
    // target object (subject to security constraints).
    //
    public DynoIndexer( object target, bool includePrivateMembers )
    {
        this.target = target;
        targetType = target.GetType();
        bindingFlags = BindingFlags.Instance | BindingFlags.Public;
        
        if( includePrivateMembers )
        {
            bindingFlags |= BindingFlags.NonPublic;
        }
    }

    public object this[ string name ]
    {
        get
        {
            MemberInfo mi = LookupMember(name);

            if( mi.MemberType == MemberTypes.Field )
            {
                // Fetch the field.
                //
                return ((FieldInfo)mi).GetValue(target);
            }
            else if( mi.MemberType == MemberTypes.Property )
            {
                // Invoke the property getter.
                //
                return ((PropertyInfo)mi).GetValue(target, null);
            }
            else if( mi.MemberType == MemberTypes.Method )
            {
                // If there is an instance method whose name matches the
                // indexer name, and that takes no parameters, and
                // whose return type exactly matches the type of the
                // "value" passed to the indexer, then invoke the
                // specified method, passing its return value back to the
                // caller.
                //
                MethodInfo methodInfo = (MethodInfo)mi;

                if( !methodInfo.IsStatic &&
                    (methodInfo.ReturnType != typeof(void)) &&
                    (methodInfo.GetParameters().Length == 0) )
                {
                    return methodInfo.Invoke(target, null);
                }
                
                throw new ArgumentException(string.Format("Specified method ({0}) is not callable.", name));
            }
            else
            {
                throw new ArgumentException("Specified member is not a field or property.", name);
            }
        }

        set
        {
            MemberInfo mi = LookupMember(name);

            if( mi.MemberType == MemberTypes.Field )
            {
                // Modify the field.
                //
                ((FieldInfo)mi).SetValue(target, value);
            }
            else if( mi.MemberType == MemberTypes.Property )
            {
                // Invoke the property setter.
                //
                ((PropertyInfo)mi).SetValue(target, value, null);
            }
            else if( mi.MemberType == MemberTypes.Method )
            {
                // This one's questionable :-)....
                //
                // If there is an instance method whose name matches the
                // indexer name, and that takes only one parameter
                // whose type exactly matches the type of the
                // "value" passed to the indexer, then invoke the
                // specified method with the specified parameter value.
                //
                MethodInfo methodInfo = (MethodInfo)mi;
                ParameterInfo[] pi = methodInfo.GetParameters();

                if( !methodInfo.IsStatic &&
                    (pi.Length == 1) &&
                    (pi[0].ParameterType == value.GetType()) )
                {
                    methodInfo.Invoke(target, new object[]{value});
                    return;
                }
                
                throw new ArgumentException(string.Format("Specified method ({0}) is not callable.", name));
            }
            else
            {
                throw new ArgumentException("Specified member is not a field or property.", name);
            }
        }
    }

    private MemberInfo LookupMember( string name )
    {
        MemberInfo[] mi = targetType.GetMember(name, bindingFlags);
        
        if( (mi == null) || (mi.Length == 0) )
        {
            throw new ArgumentException("Specified member could not be found.", name);
        }

        if( mi.Length != 1 )
        {
            throw new AmbiguousMatchException(string.Format("More than one member named '{0}' exists on target.", name));
        }

        return(mi[0]);
    }
}
